Unit RabinKarpAlg;

//////////////////////////////////////////
(*
 Hlavni myslenkou tohoto algoritmu je
 predpoklad (pomerne pravdivy), ze porovnavani
 retezcu je pomala aoperace, tudiz by se melo
 delat co nejmene. Algoritmus obsahuje specialni
 mechanismus, diky kteremu porovnava vzorek s podretezcem,
 az kdyz je velmi vysoka pravdepodobnost shody.
 Tim se pocet porovnavani muze velmi snizit.
 Zakladni verze algoritmu vsak umoznuje najednou
 vyhledavat pouze jeden vzorek.

 Necht D je velikost abecedy, M delka prohledavaneho
 textu, N delka vzorku. Pismenem T oznacme
 prohledavany text.

 Princip spociva ve vytvoreni tzv. charakteristiky
 vzorku. Ze vzorku je vytvoreno cislo o zakladu D
 pomoci Hornerova schematu, kazdy znak je jednou cifrou.
 Potom se prochazi text, vezme se podretezec, spocita
 se jeho charakteristika a dojde k jejich porovnani.
 Pokud se charakteristiky shoduji, mezi podretezcem
 a vzorkem doslo ke shode.

 Vyhodou je, ze z charakteristiky podretezce T[I..O+N]
 dokazeme v konstantnim case vypocitat charakteristiku
 retezce T[I+1...I+N+1] a to takto:

 CH(T[I+1...I+N])=(CH(T[I...I+N-1])-D*(N-1)*T[I])*D+T[I+N]

 kde CH je funkce, co vraci charakteristiku daneho podretezce.

 Dobre je, ze hodnotu D^(N-1) muzeme pocitat jen jednou a pak
 ji jen dosazovat (toto cislo budeme dale oznacovat jako H).

 Bohuzel, nic neni tak krasne, jak by se mohlo zdat. Charakteristika,
 jak jsme ji tu zavedli, muze byt pekne velke cislo; radu D^N. Proto
 se tato charakteristika jeste moduli cislem Q, ktere muzeme libovolne
 zvolit. Q se voli tak, aby splnovalo nasledujici podminky:
 + bylo prvocislem
 + D*Q se veslo do slova dane pocitacove architektury

 Duvodem prvni podminky je to, ze po provedeni operace mdoulo uz neplati,
 ze dva retezce se stejnou charakteristikou jsou shodne. Pokud takovy
 pripad nastane, musime vzorek s podretezcem porovnat. Modulo prvocislem
 zajisti lepsi rozlozeni charakteristik v rozsahu {0,1,...,Q-1}. Q se voli
 co nejvetsi z duvodu vytvoreni co nejvetsiho rozsahu pro charakteristiky.

 Po zavedeni modulo Q do charakteristiky bude vypadat nas vzorecek trosku
 jinak:

 H=D^(N-1) mod Q
 CH(T[I+1...I+N])=((CH(T[I...I+N-1])-H*T[I])*D+T[I+N]) mod Q

 *)
//////////////////////////////////////////

// TODO: Bloom filter nebo neco podobneho
// na rychle porovnavani K vzorku stejne
// delky

Interface

Uses
    TextSearcAlg;

Type
    TRabinKarpAlgorithm = Class (TTextSearchAlgorithm)
                          Private
                           FQ : Cardinal;
                           FH : Cardinal;
                           FD : Cardinal;
                           Procedure RabinKarp(APatternIndex:Integer);
                           Function  GetFirstLowerPrime(ANumber:Cardinal):Cardinal;
                           Procedure ComputeHValue(APattern:String);
                          Public
                           Procedure DoSearch; Override;
                          end;

Implementation

Function  TRabinKarpAlgorithm.GetFirstLowerPrime(ANumber:Cardinal):Cardinal;
(*
 Najde nejblizsi mensi prvocislo k zadanemu
 cislu.
 ANumber - zadane cislo
*)

// Potencialni prvocislo
Var P : Cardinal;
// Aktualni nasobek 6ti
    N : Cardinal;
    JePrvocislo : Boolean;
begin
Result:=0;
// Prohledavame vsechna nizsi cisla
For P:=ANumber-1 Downto 1 Do
    begin
    If (P Mod 2 = 0) Or (P Mod 3 = 0) Then
       Continue;
    JePrvocislo:=True;
    For N:=1 To Trunc(sqrt(P)) Div 6 Do
        begin
        If (P Mod (6*N+1) = 0) Or (P Mod (6*N-1) = 0) Then
           begin
           JePrvocislo:=False;
           Break;
           end;
        end;
    If JePrvocislo Then
       begin
       Result:=P;
       Break;
       end;
    end;
end;

Procedure TRabinKarpAlgorithm.RabinKarp(APatternIndex:Integer);
(*
 Provede vlastni Rabin-Karpuv algoritmus
 APatternIndex - index vzorku
*)
Var M : Cardinal;
    N : Cardinal;
    I : Cardinal;
    // Charakteristika vzorku
    CH : Cardinal;
    // Charakteristika podretezce
    CHT : Cardinal;
begin
// Inicializace
M:=Length(FText);
N:=Length(FPatterns[APatternIndex]);
CH:=0;
CHT:=0;
// Ziskani charakteristiky vzorku a prvniho
// podretezce
For I:=1 To N Do
    begin
    CH:=(CH*FD+FAbeceda[Ord(FPatterns[APatternIndex][I])]) Mod FQ;
    CHT:=(CHT*FD+FAbeceda[Ord(FText[I])]) Mod FQ;
    end;
For I:=1 To M-N Do
    begin
    IF CH = CHT Then
       begin
       // HIT! Musime porovnat
       // (take by slo udrzovat treba jeste jednu charakteristiku)
       If Copy(FText,I,N) = FPatterns[APatternIndex] Then
          OutputAdd(I,APatternIndex);
       end;
    // Prepocet charakteristiky v konst. case
    If I < M-N Then
       // Vyraz musime trochu upravit, jinak by dochazelo k
       // podteceni... modulo je zradne
       // Problem je zvlaste s nejvnitrnejsi zavorkou
       // jelikoz muze byt max. 3Q, tak
       // Q musi byt mensi nez MaxInt / (3*D)
       CHT:= (FD*(FQ+CHT - FH*FAbeceda[Ord(FText[I])] Mod FQ) + FAbeceda[Ord(FText[I+N])])  Mod FQ;
    end;
end;


Procedure TRabinKarpAlgorithm.ComputeHValue(APattern:String);
(*
 Vypocita velikost cisla H.
 APattern - vzorek, pro nejz ma byt H vypocitano
*)
Var I : Cardinal;
    N : Cardinal;
begin
N:=Length(APattern)-1;
If N = 0 Then
   FH:=1
Else begin
     FH:=FD;
     For I:=N Downto 2 Do
           FH:=(FH*FD) Mod FQ;
     end;
end;


Procedure TRabinKarpAlgorithm.DoSearch;
Var I : Integer;
begin
FD:=FVelikostAbecedy;
FQ:=GetFirstLowerPrime($7FFFFFFF Div (3*FD));
For I:=0 To FPatterns.Count-1 Do
    begin
    ComputeHValue(FPatterns[I]);
    RabinKarp(I);
    end;
end;

End.
