Unit TextSearcAlg;

//////////////////////////////////////////
(*
 Tato unita obsahuje "implementaci" obecneho
 algoritmu pro vyhledavani textu. Vstupem
 algoritmu je text a vzorky, jenz se maji
 vyhledavat, vystupem seznam usporadanych
 dvojic (i, v), kde i je index prvniho
 pismena vyskytu v textu a v je vzorek.

 Obecny algoritmus je implementovat najko
 abstraktni trida. Pro doprogramovani
 specifickeho algoritmu je treba prepsat
 abstraktni metodu DoSearch, jenz provede
 samotne vyhledavani.

 POZOR - podporovany jsou jen abecedy
 jednobajtove. Takze pokud chcete napr.
 vyhledavat v UNICODE textech, tak bud
 si najdete/napiste neco jineho, nebo
 lehce modifikujte kod a tvarte se, ze
 vase znaky maji porad jeden byte.
*)
//////////////////////////////////////////

Interface

Uses
      // Unita Classes obsahuje nektere
      // uzitecne tridy, jako treba
      // StringList, ktery umi slusne
      // reprezentovat seznam retezcu,
      // coz byuzivam na uchovani seznamu
      // vzorku.
      Classes;

Const
      // Maximalni podporovana velikost
      // abecedy
      MAX_ABECEDA = 256;
      NENI_V_ABECEDE = $FFFFFFFF;

Type
      // Format vystupu. Jde o usporadanou
      // dvojici (i, v). Viz. uvodni komentar.
      TOutput = Record
                // Index v prohledavanem textu
                Index : Cardinal;
                // Index vzorku v seznamu vzorku
                PatternIndex : Cardinal;
                end;
      // Takto v Delphi vypada dynamicke
      // pole.
      TOutputs = Array Of TOutput;

      // Trida reprezentujici obecny
      // algoritmus pro vyhledavani v
      // textu.
      TTextSearchAlgorithm = Class
                             Private
                              // Seznam vystupu
                              FOutputs : TOutputs;
                             Protected
                              // Abeceda tvorena znaky vstupniho textu
                              // a vzorku. Umoznuje rychle indexovani.
                              FAbeceda : Array [0..MAX_ABECEDA-1] Of Cardinal;
                              FVelikostAbecedy : Cardinal;
                              // Text, ve kterem se bude
                              // vyhledavat
                              FText     : String;
                              // Vzorky, ktere se budou
                              // vyhledavat. TStringList
                              // je v podstate pole
                              // retezcu variabilni delky.
                              FPatterns : TStringList;
                              // Jmeno algoritmu, pro ruzne
                              // vypisovaci ucely
                              FAlgName  : String;
                             Public
                              // Zjisti velikost abecedy a jakymi
                              // znaky je tvorena.
                              Procedure UrciAbecedu;
                              // Tato metoda provadi vlastni vyhledavani
                              // vzorku v textu. Jeji implementaci je
                              // dosazeno konkretniho algoritmu
                              Procedure DoSearch; Virtual; Abstract;
                              // Vytvori instanci tridy obecneho
                              // algoritmu pro vyhledavani v textu.
                              Constructor Create(AText:String; APatterns:TStringList; AAlgName:String); Virtual;
                              // Zrusi instanci algoritmu
                              Destructor Destroy; Override;
                              // Prida polozku do seznamu vystupu
                              Procedure OutputAdd(ATextIndex:Cardinal; APatternIndex:Cardinal);
                              // Ostatni objekty nebudou moci atribut
                              // (F)Outputs menit, ale pouze cist
                              Property Outputs : TOutputs Read FOutputs;
                              // Nazev algoritmu si muze kdokoliv precist,
                              // ale nemuze jej zmenit.
                              Property AlgName : String Read FAlgName;
                             end;

Implementation

Constructor TTextSearchAlgorithm.Create(AText:String; APatterns:TStringList; AAlgName:String);
(*
 Vytvori instanci obecneho algoritmu
 vyhledavajiciho v textu.

 AText - text, ve kterem se bude vyhledavat.
 APatterns - co se bude vyhledavat (vzorky)
 AAlgName - nazev algoritmu (pro ucely tisku)
 *)
begin
// Zavolame konstruktor predka. Mozna to
// neni treba, protoze to Delphi za nas
// udela samo, ale budme peclivi
Inherited Create;
// Dosadime do atributu tridy
FText:=AText;
FPatterns:=APatterns;
FAlgName:=AAlgName;
// Dalsi inicializace
SetLength(FOutputs,0);
// Zjisti mnozinu znaku, ktera tvori prohledavany
// text a vzorky.
UrciAbecedu;
end;

Destructor TTextSearchAlgorithm.Destroy;
begin
// Uklid pouzivanych prostredku
SetLength(FOutputs,0);
FText:='';
// Zavolame destruktor predka, ac to mozna
// neni treba.
Inherited Destroy;
end;

Procedure TTextSearchAlgorithm.UrciAbecedu;
(*
 Zjisti velikost abecedy a jakymi znaky
 je potreba. Umozni tak napriklad prevadet
 jednotlive znaky na cisla, coz je u nekterych
 algoritmu dulezite.
*)
Var
    // Promenne pro FOR-cykly
    I, J : Integer;
    // Do tohoto pole se bude indexovat
    // ordinalni hodnotou znaku. Pokud
    // Jej abeceda jiz obsahuje, bude
    // na danem indexu prvek s hodnotou
    // True, jinak False.
    PritomneZnaky : Array [0..MAX_ABECEDA-1] Of Boolean;
begin
// Provedeme inicializaci
FVelikostAbecedy:=0;
For I:=0 To High(FAbeceda) Do
    begin
    FAbeceda[I]:=NENI_V_ABECEDE;
    PritomneZnaky[I]:=False;
    end;
// Projdeme cely text a najdeme mnozinu
// znaku, kterou je tvoren
For I:=1 To Length(FText) Do
    begin
    If Not PritomneZnaky[Ord(FText[I])] Then
       begin
       // I-ty znak textu jeste neni v
       // nasi abecede. Pridame ho tam.
       PritomneZnaky[Ord(FText[I])]:=True;
       FAbeceda[Ord(FText[I])]:=FVelikostAbecedy;
       Inc(FVelikostAbecedy);
       end;
    end;
// Musime projit take seznam vzorku, protoze
// ve vzorcich se mohou vyskytovat znaky,
// ktere se v prohledavanem textu nevyskytuji
For J:=0 To FPatterns.Count-1 Do
For I:=1 To Length(FPatterns[J]) Do
    begin
    If Not PritomneZnaky[Ord(FPatterns[J][I])] Then
       begin
       // I-ty znak J-teho vzorku jeste
       // neni v nasi abecede. Pridame ho
       // tam.
       PritomneZnaky[Ord(FPatterns[J][I])]:=True;
       FAbeceda[Ord(FPatterns[J][I])]:=FVelikostAbecedy;
       Inc(FVelikostAbecedy);
       end;
    end;
end;

Procedure TTextSearchAlgorithm.OutputAdd(ATextIndex:Cardinal; APatternIndex:Cardinal);
(*
 Prida polozku do seznamu vystupu.

 ATextIndex - pozice prvniho znaku vyskytu vzorku
 APatternIndex - index vzorku v seznamu vzorku
*)
begin
SetLength(FOutputs,Length(FOutputs)+1);
FOutputs[High(FOutputs)].Index:=ATextIndex;
FOutputs[High(FOutputs)].PatternIndex:=APatternIndex;;
end;

End.
